
#include <admodel.h>
#include <amp.h>

dmatrix make_dmatrix(int n,int m)
{
  dmatrix M(1,n);
  dvector v(1,n*m);
  int offset=0;
  for (int i=1;i<=n;i++)
  {
    M(i)=v(1+offset,m+offset).shift(1);
    offset+=m;
  }
  return M;
}

//----------------------------------------------------------------------------
// Implement simple matrix multiplication on GPU using C++ AMP
//----------------------------------------------------------------------------
dmatrix mxm_amp_simple(const dmatrix & va, const dmatrix&  vb,dmatrix & vc) 
{
  int m=va.indexmax();
  int n=va(m).indexmax();
  int w1=vb.indexmax();
  cout << "vb.indexmax()= " << vb.indexmax() << endl;
  int w=vb(w1).indexmax();
  if (n!=vb.indexmax())
  {
    cerr << "size error in mpy" << endl;
    exit(1);
  }

    concurrency::extent<2> e_a(m, n), e_b(n, w), e_c(m, w);

    // Copy in
    concurrency::array_view<const double, 2> av_a(e_a, va); 
    concurrency::array_view<const double, 2> av_b(e_b, vb); 
    concurrency::array_view<double, 2> av_c(e_c, vc);
    av_c.discard_data();

    // Compute - outer 2 for loops of CPU is replaced by a parallel_for_each
    parallel_for_each(av_c.get_extent(), [=](concurrency::index<2> idx) restrict(amp)
    {
        double result = 0;

        for(int i = 0; i < av_a.get_extent()[1]; ++i)
        {
            concurrency::index<2> idx_a(idx[0], i);
            concurrency::index<2> idx_b(i, idx[1]);

            result += av_a[idx_a] * av_b[idx_b];
        }

        av_c[idx] = result;
    });
    // explicitly about copying out data
    av_c.synchronize();
    return vc;
}



  void ad_boundf(int i)
  {
    /* so we can stop here */
    exit(i);
  }
      

int  main()
{
  ad_set_new_handler();
  ad_exit=&ad_boundf;

  int n=2000;   // we will work with an nxn symmetric matrix
  int m=2000;   // we will work with an nxn symmetric matrix
  int k=2000;   // we will work with an nxn symmetric matrix
  dvector v1(1,n*k);
  dvector v2(1,k*m);
  dvector v3(1,n*m);
  // make dmatrices whose memory layiout is contiguous
  dmatrix M=make_dmatrix(n,k);
  dmatrix N=make_dmatrix(k,m);
  dmatrix P=make_dmatrix(n,m);
  random_number_generator rng(135);
  M.fill_randn(rng);  // fill M with random numbers 
  N.fill_randn(rng);  // fill M with random numbers 
  
  adtimer adt;
  mxm_amp_simple(M,N,P);
  cout << "time for AMP matrix multiply " 
       << adt.get_elapsed_time_and_reset() << endl;
  cout << "should print 0 if matrix multiply works" << endl;
  cout << norm2(P-M*N) << endl;
  cout << "time for ADMB matrix multiply " 
       << adt.get_elapsed_time_and_reset() << endl;
}
